/*
 * ioGame
 * Copyright (C) 2021 - 2023  渔民小镇 （262610965@qq.com、luoyizhu@gmail.com） . All Rights Reserved.
 * # iohao.com . 渔民小镇
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
package com.iohao.mmo.equip.service;

import com.iohao.mmo.common.config.GameCode;
import com.iohao.mmo.equip.entity.BasicEquipProperty;
import com.iohao.mmo.equip.entity.Equip;
import com.iohao.mmo.equip.entity.Goods;
import com.iohao.mmo.equip.utils.RandomUtils;
import com.mongodb.bulk.BulkWriteResult;
import lombok.AllArgsConstructor;
import org.springframework.data.mongodb.core.BulkOperations;
import org.springframework.data.mongodb.core.MongoTemplate;
import org.springframework.data.mongodb.core.query.Criteria;
import org.springframework.data.mongodb.core.query.Query;
import org.springframework.data.mongodb.core.query.Update;
import org.springframework.stereotype.Service;

import java.math.BigDecimal;
import java.util.Arrays;
import java.util.List;
import java.util.Optional;

/**
 * @author 唐斌
 * @date 2023-07-30
 * @description: 装备属性实现类
 */
@Service
@AllArgsConstructor
public class EquipService {
    final MongoTemplate mongoTemplate;

    public Equip findById(String id) {
        return  mongoTemplate.findById(id, Equip.class);
    }

    /**
     * 根据用户查找所有装备
     * @param userId
     * @param store 所属位置：1包裹，2仓库，3穿戴
     * @return
     */
    public List<Equip> listByUser(long userId,int store) {
        Query query = new Query();
        query.addCriteria(Criteria.where("userId").is(userId));
        query.addCriteria(Criteria.where("delFlag").is(false));
        if(store>0) {
            query.addCriteria(Criteria.where("goods.store").is(store));
        }

        return  mongoTemplate.find(query,Equip.class);
    }

    public void delBatch(String ids) {
        String[] idsArr = ids.split("-");
        GameCode.objNotFound.assertTrue(idsArr.length>0);
        Query query = new Query();
        query.addCriteria(Criteria.where("id").in(Arrays.asList(idsArr)));

        // 更新的字段
        Update update = new Update();
        //获取分配的字段
        update.set("delFlag", true);
        update.set("goods.delFlag", true);

        //UNORDERED是平行处理，即使某条记录出错了，其余的也会继续处理；ORDERED是队列排序处理，只要中途有个失败了，那么后续的操作流程就会终止了。
        BulkOperations bulkOps = mongoTemplate.bulkOps(BulkOperations.BulkMode.UNORDERED, Equip.class);
        bulkOps.updateMulti(query, update);
        BulkWriteResult result = bulkOps.execute();
        //逻辑删除对应的商品
        delGoodsByChildBatch(idsArr);
    }

    /**
     * 根据子表id批量删除物品（逻辑删除）
     * @param detailIdArr 子表主键
     */
    public void delGoodsByChildBatch(String[] detailIdArr) {
        if (detailIdArr.length==0) {
            return;
        }
        Query query = new Query();
        query.addCriteria(Criteria.where("detailId").in(Arrays.asList(detailIdArr)));

        // 更新的字段
        Update update = new Update();
        //获取分配的字段
        update.set("delFlag", true);

        //UNORDERED是平行处理，即使某条记录出错了，其余的也会继续处理；ORDERED是队列排序处理，只要中途有个失败了，那么后续的操作流程就会终止了。
        BulkOperations bulkOps = mongoTemplate.bulkOps(BulkOperations.BulkMode.UNORDERED, Goods.class);
        bulkOps.updateMulti(query, update);
        BulkWriteResult result = bulkOps.execute();
    }

    /**
     * 分配装备属性点
     * @param equip
     */
    public void allotEquip(Equip equip){
        //判定总点数是否合法
        Equip oldEquip = mongoTemplate.findById(equip.getId(), Equip.class);

        GameCode.objNotFound.assertTrue(oldEquip!=null);

        BasicEquipProperty basicEquipProperty = equip.getBasicEquipProperty();
        int saveAttrTotal = basicEquipProperty.getConstitution()+basicEquipProperty.getMagicPower()+basicEquipProperty.getPower()+basicEquipProperty.getEndurance()+basicEquipProperty.getAgile();

        GameCode.allotNotEnough.assertTrue(saveAttrTotal<=oldEquip.getAttrTotal());

        // 查询条件，如果数据存在更新
        Query query = new Query();
        query.addCriteria(Criteria.where("id").is(equip.getId()));

        // 更新的字段
        Update update = new Update();
        //获取分配的字段

        update.set("undistributedAttr", oldEquip.getAttrTotal()-saveAttrTotal);
        update.set("basicEquipProperty", basicEquipProperty);
        mongoTemplate.upsert(query,update,Equip.class);
    }

    /**
     * 重新随机总属性点
     * @param itemId 装备物品id
     * @param excellentRate 增加的极品率
     */
    public Equip resetEquip(String itemId, BigDecimal excellentRate){
        Equip equip = findById(itemId);
        GameCode.objNotFound.assertTrue(equip!=null);

        int randomMin = equip.getTotalAttrMin();
        int randomMax = equip.getTotalAttrMax();
        //品质
        boolean excellentFlag = RandomUtils.isExcellent(excellentRate);//是否生成极品
        equip.setQuality(excellentFlag?2:1);
        //随机属性值
        int newAttrTotal = RandomUtils.randomFromExcellent(randomMin,randomMax,excellentFlag);
        // 直接替换掉原装备的属性值
        return replaceEquipAttr(equip,newAttrTotal);
    }

    /**
     * 替换掉原来装备的属性值，并且重置已分配的属性点
     * @param equip
     * @param newAttrTotal
     * @return
     */
    private Equip replaceEquipAttr(Equip equip,int newAttrTotal){
        equip.setBasicEquipProperty(BasicEquipProperty.resetBasicEquipProperty());
        equip.setAttrTotal(newAttrTotal);
        equip.setUndistributedAttr(newAttrTotal);
        mongoTemplate.save(equip);
        return equip;
    }

}
